#include <linux/init.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/platform_device.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <asm/io.h>
#include <linux/slab.h>
#include <linux/gpio.h>

/*******************都是真实的物理地址********************/
/******************************************************/
#define SYS_BA              0xb0000000      //SYS Base Address
//key enable pin
#define SYS_GPB_MFPL        (SYS_BA+0x78)   //GPIOB Low Byte Multiple Function Control Register
#define SYS_GPB_MFPH        (SYS_BA+0x7C)   //GPIOB High Byte Multiple Function Control Register

/******************************************************/
#define GPIO_BA             0xb8003000      //GPIO Base Address:
//key
#define GPIOB_DIR           (GPIO_BA+0x40)  //GPIO Port B Direction Control Register
#define GPIOB_DATAOUT       (GPIO_BA+0x44)  //GPIO Port B Data Output Register
#define GPIOB_DATAIN        (GPIO_BA+0x48)  //GPIO Port B Data Outin Register   //qianxuping, key input data reg

/******************************************************/
#define CLK_BA              0xb0000200      //CLK Base Address
#define CLK_PCLKEN0         (CLK_BA+0x18)   //APB Devices Clock Enable Control Register 0

#define REGISTER_LENGTH	    4

void key_dev_release(struct device *dev)
{
    printk("%s is called\n", __func__);
}

struct resource key_resource[] =
{
  [0] = {
    .start = SYS_GPB_MFPL,
    .end = SYS_GPB_MFPL+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
  [1] = {
    .start = SYS_GPB_MFPH,
    .end = SYS_GPB_MFPH+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
  [2] = {
    .start = GPIOB_DIR,
    .end = GPIOB_DIR+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
  [3] = {
    .start = GPIOB_DATAOUT,
    .end = GPIOB_DATAOUT+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
  [4] = {
    .start = GPIOB_DATAIN,
    .end = GPIOB_DATAIN+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
  [5] = {
    .start = CLK_PCLKEN0,
    .end = CLK_PCLKEN0+REGISTER_LENGTH-1,
    .flags = IORESOURCE_MEM,
  },
};

struct platform_device key_dev =
{
    .name = "nuc977-key",
    .id = -1,
    //.num_resources = 1,
    .num_resources = ARRAY_SIZE(key_resource), //直接通过这个求出数组元素的个数
    .resource = key_resource,
    .dev={
        .release = key_dev_release,
    },
};

int key_dev_init(void)
{
  printk("key_dev_init\r\n");

  platform_device_register(&key_dev);

  return 0;
}

void key_dev_exit(void)
{  
  printk("key_dev_exit\r\n");

  platform_device_unregister(&key_dev);
}

module_init(key_dev_init);
module_exit(key_dev_exit);
MODULE_AUTHOR("qianxuping");
MODULE_LICENSE("GPL");

