#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <asm/io.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/gpio.h>
#include <mach/irqs.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/mutex.h>
#include <linux/input.h>
#include <linux/platform_device.h>
#include <linux/mod_devicetable.h>
#include <linux/pm_wakeup.h>
#include <linux/pm.h>

#define	NUC970_PB11	(0x20 + 11)

#define KEY_NUM 1
#define KEY_OFF 0
#define KEY_ON  1

/* gpio虚拟地址指针 */
/*********key gpio is PB0,PB1***********/
static unsigned long *gpbmfpl = NULL;		//复用寄存器低32位
static unsigned long *gpbmfph = NULL;		//复用寄存器高32位
static unsigned long *gpbdir = NULL;		//方向寄存器
static unsigned long *gpbdataout = NULL;	//输出寄存器
static unsigned long *gpbdatain = NULL;   	//输如寄存器
static unsigned long *pclken0 = NULL;		//时钟寄存器
// static void __iomem *gpbmfpl = NULL;		//复用寄存器低32位
// static void __iomem *gpbmfph = NULL;		//复用寄存器高32位
// static void __iomem *gpbdir = NULL;		//方向寄存器
// static void __iomem *gpbdataout = NULL;	//输出寄存器
// static void __iomem *gpbdatain = NULL;   	//输如寄存器
// static void __iomem *pclken0 = NULL;		//时钟寄存器

const char key_dev_name[]="wj-key";
struct key_dev{
    dev_t devid;
    struct class     *cls;
    struct device    *dev[KEY_NUM];
    int major;
    int minor;
};

struct key_dev g_key_dev;

DECLARE_WAIT_QUEUE_HEAD(button_wq);
static volatile int ev_press = 0;

struct timer_list timer;
void *my_dev_id = NULL;

struct key_desc
{
  int key_pin;
  int key_val;
};

struct key_desc key_table[KEY_NUM] = 
{
  {NUC970_PB11, 0x3},//release->0x3, press->0x83
};

static unsigned int interval_timer=300;//定时间隔，设置成变量，方便用户空间可以修改这个间隔

static unsigned char is_dev_sleep_or_run = 0;

//static unsigned char which_irq[KEY_NUM] = {0}; //这样写还不能产生中断
static int which_irq[KEY_NUM] = {0}; //记录各个按键申请到的中断号

//static struct wakeup_source *key_wakeup_ws;

irqreturn_t key_irq_handler(int irq, void *dev_id)
{
    //按键中断处理函数中，开启定时器

    printk("key_irq_handler\r\n");

    if(2 == is_dev_sleep_or_run) //在睡眠状态，按下按键，则唤醒设备
    {
        is_dev_sleep_or_run = 1;

        //__pm_stay_awake(key_wakeup_ws);
        //__pm_relax(key_wakeup_ws);

        //方法1
        //pm_stay_awake(g_key_dev.dev[0]);
        //pm_relax(g_key_dev.dev[0]);

        //方法2
        pm_wakeup_event(g_key_dev.dev[0], 0);

        printk("key is down: 按下按键, 唤醒设备\r\n");
    }
    else //不是睡眠状态则添加定时器，进入3s是否是按下判断流程
    {
        //mod_timer(&timer, jiffies+HZ/50);   //修改定时器的超时时间,100代表1s，这个只有0.02s？
        mod_timer(&timer, jiffies+interval_timer);   //修改定时器的超时时间,100代表1s
        //my_dev_id = dev_id;  //这样会崩溃
        my_dev_id = (struct key_desc *)dev_id;
    }
    
    return IRQ_HANDLED;
}

/******************************flies************************************/
static int key_open(struct inode *inode, struct file *filp)
{
    int i = 0;
    int ret = 0;

	filp->private_data = &g_key_dev;  //private data是个void *类型的数据，相当于是个万能指针,这样以后操作filp->private_data就相当于操作g_key_dev

	printk("%s is called\n", __func__);

    *pclken0 |= (1<<3);                     //APB Devices Clock Enable,这里用了，在porbe的时候，一定要先iomap，获取到地址，不然会崩溃
    *gpbmfph &= ~(0x0f<<12);                //Multiple Function,PB11通用gpio功能 ,这里用了，在porbe的时候，一定要先iomap，获取到地址，不然会崩溃
    *gpbdir &= ~(0x1 << 11);                //我是11脚,PB11,方向为输入,这里用了，在porbe的时候，一定要先iomap，获取到地址，不然会崩溃

#if 1 //在porble函数里申请
    //申请中断
    for (i = 0; i < sizeof(key_table)/sizeof(key_table[0]); i++)
    {
        //ret = request_irq(gpio_to_irq(key_table[i].key_pin), key_irq_handler, IRQF_TRIGGER_FALLING|IRQF_TRIGGER_RISING, "lowpower-key", &key_table[i]);
        //which_irq[i]=gpio_to_irq(key_table[i].key_pin);
        printk("which_irq[%d] is %d\n", i, which_irq[i]);

        ret = request_irq(which_irq[i]=gpio_to_irq(key_table[i].key_pin), key_irq_handler, IRQF_TRIGGER_FALLING, "wj-key-irq", &key_table[i]);
        if (ret)
            printk("request_irq %d failed, ret = %d\n", i, ret);
    }

#endif

    //key_wakeup_ws = wakeup_source_register("keysleep");  

	return 0;
}

static ssize_t key_read(struct file *filp, char __user *buf, size_t cnt, loff_t *offt)
{
	printk("key_read\r\n");

    wait_event_interruptible(button_wq, ev_press);
    ev_press = 0;

    copy_to_user(buf, &key_table->key_val, 1);

	return 0;
}

static ssize_t key_write(struct file *filp, const char __user *buf, size_t cnt, loff_t *offt)
{
	/*写成功后，返回写入的字数*/
	return 0;
}

static int key_release(struct inode *inode, struct file *filp)
{
    int i = 0;

	printk("key_release\r\n");

    for (i = 0; i < sizeof(key_table)/sizeof(key_table[0]); i++)
    {
        free_irq(gpio_to_irq(key_table[i].key_pin), &key_table[i]);
    }

	return 0;
}

static struct file_operations key_fops = {
	.owner = THIS_MODULE,
	.open = key_open,
	.read = key_read,
	.write = key_write,
	.release = 	key_release,
};

void my_timer_function(unsigned long data)
{
    unsigned int reg_val;
    int shift_num = 0;
    
    struct key_desc *key_desc = (struct key_desc *)my_dev_id;
    
    reg_val = *gpbdatain;   //在定时器里，3s后在读取按键，一直为高电平，则进入休眠
    
    shift_num = (key_desc->key_pin == NUC970_PB11) ? 11:11;

    if (reg_val & (1<<shift_num)) //判断是否为高电平
    {
        key_desc->key_val &= ~(1<<7);  //0x03&(0x01<<7),将最高位置0

        printk("key is down: 按下按键, 3s后判断,此时是高电平（说明在3s内松开了按键）r\n");
    }
    else
    {
        is_dev_sleep_or_run = 2;
        key_desc->key_val |= (1<<7);    //将最高位置1，相当于是刚好为0x80再加3
        
        printk("key is down: 按下按键, 3s后判断,此时还是低电平（说明按键一直按着）,设备进入睡眠\r\n");
    }
    
    printk("key_desc->key_val=0x%x\n", key_desc->key_val);

    ev_press = 1;
    wake_up_interruptible(&button_wq);
}

/******************************platform************************************/
static int key_drv_probe(struct platform_device *pdev)
{
    int i;
    int ret;
	int key_resource_size[5];
    //struct resource *key_resource[6] = NULL; //这样写报错？
	struct resource *key_resource[6];

	printk("key_drv_probe\r\n");

    /***************add timer***************/
    init_timer(&timer);
    timer.function = my_timer_function;
    timer.data = 1; //传递给定时器超时处理函数的值
    add_timer(&timer);
    /******************************/

    g_key_dev.major = 0;

    if (g_key_dev.major != 0)
        register_chrdev(g_key_dev.major, key_dev_name, &key_fops);
    else
        g_key_dev.major = register_chrdev(0, key_dev_name, &key_fops);

    printk("g_key_dev.major=%d\n", g_key_dev.major);

    g_key_dev.cls = class_create(THIS_MODULE, key_dev_name);
    if (IS_ERR(g_key_dev.cls ))
    {
        printk("class_create failed\n");
        goto error;
    }
    
    for(i=0; i<KEY_NUM; i++)
    {
        g_key_dev.dev[i] = device_create(g_key_dev.cls , NULL, MKDEV(g_key_dev.major, i), NULL, "wj-key%d", i);
        if (IS_ERR(g_key_dev.dev[i]))
        {
            printk("wj-key%d create failed\n", i);
            goto error;
        }
    }

	for(i=0; i<6; i++)	//这个6就是在设备文件定义的key资源数组的个数
    {
		printk("i=%d\n", i);

		#if 0
		key_resource[i] = platform_get_resource(pdev, IORESOURCE_MEM, i);
		if(!key_resource[i] )
		{
			printk("platform get key_resource[%d] faikey\n", i);

			goto error;
		}
		printk("res->start=%d\n", key_resource[i]->start);
		printk("res->end=%d\n", key_resource[i]->end);

		key_resource_size[i] = resource_size(key_resource[i]);
		printk("key_resource_size[%d]=%d\n", i, key_resource[i]);
		#endif

		#if 1
		if(i<6)
		{
			key_resource[i] = platform_get_resource(pdev, IORESOURCE_MEM, i);
			if(!key_resource[i] )
			{
				printk("platform get key_resource[%d] failed\n", i);

				goto error;
			}
			printk("res->start=%x\n", key_resource[i]->start);   //-1342177160 为什么是负数
			printk("res->end=%x\n", key_resource[i]->end);
			
			key_resource_size[i] = resource_size(key_resource[i]);
			printk("key_resource_size[%d]=%x\n", i, key_resource_size[i]);
		}
		#endif
    }

#if 0 //为什么用这种iomap会崩溃，难道是因为负数的原因？
    /*****************key pin iomap***********************/
    gpbmfpl =  ioremap(key_resource[0]->start, key_resource_size[0]);
	
    //gpbmfph = gpbmfpl + 1;
	gpbmfph =  ioremap(key_resource[1]->start, key_resource_size[1]);

    gpbdir =  ioremap(key_resource[2]->start, key_resource_size[2]);

    //gpbdataout = gpbdir + 1;
	gpbdataout =  ioremap(key_resource[3]->start, key_resource_size[3]);

    //gpbdatain= gpbdataout + 1;
	gpbdatain =  ioremap(key_resource[4]->start, key_resource_size[4]);

    /*****************gpio clk iomap***************************/
    pclken0 =  ioremap(key_resource[5]->start, key_resource_size[5]);
#endif

    /*****************key pin iomap***********************/
    gpbmfpl = (unsigned long *)ioremap(0xb0000078, 16);
    gpbmfph = gpbmfpl + 1;

    gpbdir = (unsigned long *)ioremap(0xb8003040, 16);  
    gpbdataout = gpbdir + 1;
    gpbdatain= gpbdataout + 1;

    /*****************gpio clk iomap***************************/
    pclken0 = (unsigned long *)ioremap(0xb0000218, 16);

    /*****************key pin 申请中断***************************/
#if 0
    //申请中断
    for (i = 0; i < sizeof(key_table)/sizeof(key_table[0]); i++)
    {
        //ret = request_irq(gpio_to_irq(key_table[i].key_pin), key_irq_handler, IRQF_TRIGGER_FALLING|IRQF_TRIGGER_RISING, "lowpower-key", &key_table[i]);
        //which_irq[i]=gpio_to_irq(key_table[i].key_pin);
        //printk("which_irq[%d] is %d\n", i, which_irq[i]);

        ret = request_irq(which_irq[i]=gpio_to_irq(key_table[i].key_pin), key_irq_handler, IRQF_TRIGGER_FALLING, "wj-key-irq", &key_table[i]);
        if (ret)
            printk("request_irq %d failed, ret = %d\n", i, ret);

        printk("which_irq[%d] is %d\n", i, which_irq[i]);
    }
#endif

    ret=device_init_wakeup(&pdev->dev, 1);
    //ret=device_init_wakeup(&g_key_dev.dev, 1);
    if(ret<0)
        printk("device_init_wakeup set failed\r\n");

#if 1
	//这些初始化必须要放在iomap后面使用，如果放在前面，相当于还没有申请就使用指针，会崩溃
#endif

	return 0;
error:
  return -1;
}

static int key_drv_remove(struct platform_device *pdev)
{
    int i;

	printk("key_drv_remove\r\n");

    for (i=0; i<KEY_NUM; i++)
    {
        device_destroy(g_key_dev.cls , MKDEV(g_key_dev.major, i));
    }
    class_destroy(g_key_dev.cls);
    unregister_chrdev(g_key_dev.major, key_dev_name);

    iounmap(gpbmfpl);
    iounmap(gpbmfph);
    iounmap(gpbdir);
    iounmap(gpbdataout);
    iounmap(gpbdatain);

    iounmap(pclken0);

    del_timer(&timer);

    for (i = 0; i < sizeof(key_table)/sizeof(key_table[0]); i++)
    {
        free_irq(gpio_to_irq(key_table[i].key_pin),  &key_table[i]);
    }

    device_init_wakeup(&pdev->dev, 0);

	return 0;
}

//static int key_drv_suspend(struct platform_device * pdev, pm_message_t state)
static int key_drv_suspend(struct device * dev, pm_message_t state)
{
    printk("key supend\r\n");

    printk("key_drv_suspend's which_irq[0] is %d\n", which_irq[0]);

    printk("pm_message_t is %x\r\n", state.event);

    enable_irq_wake(which_irq[0]);//使能中断唤醒

    return 0;
}
	
//static int key_drv_resume(struct platform_device * pdev)
static int key_drv_resume(struct device * dev)
{
    printk("key resume\r\n");

    is_dev_sleep_or_run = 2;

    printk("key_drv_resume key is down: 按下按键, 唤醒设备\r\n");

    disable_irq_wake(which_irq[0]);//关闭中断唤醒

    return 0;
}

//static SIMPLE_DEV_PM_OPS(wj_key_pm_ops, key_drv_suspend, key_drv_resume);
static const struct dev_pm_ops wj_key_pm_ops = {
	.suspend = key_drv_suspend,
	.resume = key_drv_resume,
};

static struct platform_device_id key_drv_ids[] = {
	{.name = "nuc977-key"},//匹配的名字
	{}
};

static struct platform_driver key_driver = {
    .driver = {
        .name = "nuc977-key",
        .owner = THIS_MODULE,
        .pm = &wj_key_pm_ops,
    },
	.probe	    = key_drv_probe,
	.remove	    = key_drv_remove,
	.id_table   = key_drv_ids,
    // .suspend    = key_drv_suspend,
    // .resume     = key_drv_resume,
};

int keyh_drv_init(void)
{
    printk("keyh_drv_init\r\n");
    
    platform_driver_register(&key_driver);

    return 0;
}

void keyh_drv_exit(void)
{ 
    printk("keyh_drv_exit\r\n");

    platform_driver_unregister(&key_driver);
}

module_init(keyh_drv_init);
module_exit(keyh_drv_exit);
MODULE_AUTHOR("qianxuping");
MODULE_LICENSE("GPL");

